<?php

/*

   MultiViews
   ==========
   content-negotiation for nanoweb

   - doesn't makes regex from directoryindex[0] anymore for speed purposes
   - returns faster now for existing files
   - a faster split_path_info() is there, as most times there is no
     path_info it returns faster, no regexs more in it

   mario@erphesfurt.de

*/



class mod_multiviews {



   function mod_multiviews() {

      $this->modtype="core_after_decode";
      $this->modname="MultiViews (content negotiation)";

      $GLOBALS["mime_enc"]["gz"] = "gzip";
      $GLOBALS["mime_enc"]["Z"] = "compress";

   }




      # fetches extensions regex from global mime table 
      #

   function extensions($string) {
      global $mime;

      $regex = '';
      $string = preg_replace('/[*]/', '.+', $string);
      
      foreach ($mime as $ext => $mimetype) {
         if (preg_match("#^{$string}$#i", $mimetype)) {
            $regex .= ($regex ? "|" : "") . $ext;
         }
      }

      if (empty($regex))  return;

      return("($regex)");
   }




      # returns array with quality values ordered by
      # file name extensions

   function parseQualities($qstring, $q = 0.9) {

      $qualities = array();

      foreach (split(",", $qstring) as $qspart) {

         if (preg_match('#^\s*([^\s;]+)#i', $qspart, $uu)) {

            //-- file extension
            if (! preg_match('#/#', $uu[1])) {
               $ext = preg_replace('#_.+$#', '', $uu[1]);
            }
            else {
               $ext = $this->extensions($uu[1]);
            }

            if (! empty($ext)) {
               //-- quality value
               if (preg_match('#qs?=([.0-9]+)#i', $qspart, $uu)) {
                  $q = $qualities[$ext] = $uu[1];
               }
               else {
                 $qualities[$ext] = $q;
               }
               $q /= 1.7;
            }
         }
      }

      return($qualities);
   }





      # // returns list($http_uri,$path_info) generated from $http_uri
      # // now based on string arithmetics
      # not used anymore, as path_info now gets removed
      # in the server core right before we get invoked

   function split_path_info() {
      global $http_uri, $docroot;
      global $path_info;   // global $path_info to actually correct value

      $dir = substr($http_uri, 0, strrpos($http_uri, "/"));
      if (is_dir($docroot . $dir)) {
         return(array($http_uri, $path_info = ""));
      }  // quick

      $slash_pos = 0;
      while (($slash_pos = strpos($http_uri, "/", ++$slash_pos)) !== false) {
         $dir = substr($http_uri, 0, $slash_pos);
         if (! is_dir($docroot . $dir)) {
            break;
         }
      }

      return(array("$dir", $path_info = substr($http_uri, $slash_pos, 256)));
   }




   function main() {

      global $http_uri, $conf, $docroot, $htreq_headers, $mime_enc, $out_add_headers, $pri_err;
      if (@$conf["global"]["reflectrewriting"]) {
         global $real_uri;
      }

       if (substr($http_uri, strrpos($http_uri, "/"), 255) && file_exists($docroot . $http_uri)) {
         return("no negotiation because file exists");
       }  // quick skip1

      // determine file name to be worked on
      $path_info = "";
      $file = $http_uri;
                 //unused: list($file, $path_info) = $this->split_path_info();
      $act_path = "";
      if ($last_slash = strrchr("/", $file)) {
         $act_path = substr($file, 0, $last_slash);
         $file = array(basename($file));
      }
#DEBUG#techo("AFP == $act_path :: {$file[0]} :: $path_info");

      // add '/index.html/', because main server doesn't search variants
      if (empty($file)) { 
         $file = explode(" ", trim($conf["global"]["directoryindex"][0]));
      }

      // find files with same basename and different extensions
      $files = array();
      $dir = opendir($docroot . $act_path);
      while($filename = readdir($dir)) {
         foreach ($file as $filebn) {
            if (substr($filename, 0, strlen($filebn)) == $filebn) {
               $files["$filename"] = 0.0;
            }
         }
      }
      closedir($dir);

      if (empty($files) || file_exists($docroot . $act_path . $file)) {
         return("no rewriting occoured");
      }  // quick skip2


      // priorities
      $qualities = array_merge(
         $this->parseQualities(strtr(@$conf["global"]["languagepriority"][0], " ", ", "), 0.5),
         $this->parseQualities(@$htreq_headers["ACCEPT-LANGUAGE"], 1.0),
         array(
            "html" => 0.009,   // should we prefer .php instead??  ;-)
            "png"  => 0.003,
            "jpeg" => 0.002,
            "gif"  => 0.001    // ... these are micro-priorities
         ),
         $this->parseQualities(implode(", ", @$conf["global"]["otherpriority"]), 0.3),
         $this->parseQualities(@$htreq_headers["ACCEPT"], 1.0)
      );
      foreach ($mime_enc as $ext => $encoding) {
         if (strpos("{$encoding}", $htreq_headers["ACCEPT-ENCODING"]) !== false) {
            $qualities = array_merge($qualities, array($ext => 0.5));
         }
      }

      // go thru filename extensions, and multiply qualities
      foreach ($files as $filename => $file_q) {

         preg_match("#^{$file}(.+)$#", $filename, $uu);
         foreach (split('[.]', $uu[1]) as $ext) {

            if (! $ext) { continue; }

            foreach ($qualities as $regex => $q) {
               if (preg_match("#{$regex}#", "$ext")) {
                  $files[$filename] += $q;
               }
            }

         }
      }

      // sort
      arsort($files, SORT_NUMERIC);

      list($f0, $f1) = array_values($files);
      if ($f0 == $f1) {
         # ... 
      }

     // return "best" variant
     list($file) = array_keys($files);
     $real_uri=$http_uri = $act_path . $file . $path_info;

     // server-driven negotiation
     $negotiation_agent = strtolower(trim(@$conf["global"]["negotiation"][0])) == "agent";
     if ( ($negotiation_agent || strlen($path_info))
           && empty($htreq_headers["NEGOTIATE"]) )
     {
        $out_add_headers["Content-Location"] = $path_info ? ($act_path . $file . $path_info) : $file;
        $out_add_headers["Vary"] = "negotiate"; // which request headers affect selection
        $out_add_headers["TCN"] = "choice";

        // report encoding to client
        foreach ($mime_enc as $ext => $encoding) {
           if (strpos($file, ".{$ext}")) {
              $out_add_headers["Content-Encoding"] = $encoding;
              $GLOBALS["out_encoded"] = true;
           }
        }

     }

     // agent-driven negotiation
     else {
        techo("WARN: user agent content negotiation was selected");

        $out_add_headers["Alternates"] = implode(" , ", array_keys($files));

        $GLOBALS["pri_err"] = 300;  // HTTP 300 Choice Yourself
        $GLOBALS["out_contenttype"] = "text/html";
        $GLOBALS["lf"] = "<HTML><HEAD><TITLE>Choose Variant</TITLE></HEAD><BODY><H2>Choose Variant</H2>The document you requested exists in different variants, and your browser gives you the ability to select one of them:<BR><UL>";
        foreach ($files as $f => $q) {
           $GLOBALS["lf"] .= '<LI><A HREF="' . url_to_absolute($act_path . $f) . '">' . $f . '</A></LI>';
        }
        $GLOBALS["lf"] .= "</UL><HR>" . _genpage_signature() . "</BODY></HTML>";

     }

     // may help(?)
     if ($path_info && $act_path) {
        $out_add_headers["Content-Base"] = $act_path;
     }

#foreach ($qualities as $e => $q) techo("qualities: '.$e' => $q");
#foreach ($files as $e => $q) techo("files: '$e' => $q");

   }

}

?>